"""Medical NLP data and models"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/65_medical.text.ipynb.

# %% ../../nbs/65_medical.text.ipynb 4
from __future__ import annotations
from ..basics import *

# %% auto 0
__all__ = []
